/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication;

import java.nio.ByteBuffer;
import org.jflux.api.core.Adapter;
import org.jflux.api.core.Listener;
import org.jflux.api.core.chain.AdapterChain;
import org.jflux.api.core.chain.ListenerChain;
import org.jflux.api.core.event.Event;
import org.jflux.api.core.event.Header;
import org.jflux.api.core.event.Header.HeaderPropertyAdapter;
import org.jflux.api.core.node.DefaultConsumerNode;
import org.jflux.api.data.routing.Router.DefaultRouter;
import org.jflux.impl.encode.avro.SerializationConfigUtils;

/**
 *
 * @author Matthew Stevenson
 */
public class RoutingDecoderNode extends 
        DefaultConsumerNode<Event<Header,ByteBuffer>>{
    private DefaultRouter<String,Event<Header,ByteBuffer>> myDecodingRouter;
    private DefaultRouter<String,Event<Header,?>> myEventRouter;
    
    public static RoutingDecoderNode makeDefault(){
        return new RoutingDecoderNode(
                new DefaultRouter<String,Event<Header,ByteBuffer>>(
                new AdapterChain(
                        new Event.EventHeaderAdapter(),
                        new HeaderPropertyAdapter(
                                SerializationConfigUtils.CONF_MESSAGE_CLASS))));
    }
    
    RoutingDecoderNode(
            DefaultRouter<String,Event<Header,ByteBuffer>> decoder){
        super(decoder);
        myDecodingRouter = decoder;
        myEventRouter = new DefaultRouter<String, Event<Header, ?>>(
                new AdapterChain(
                        new Event.EventHeaderAdapter(),
                        new Header.HeaderTypeAdapter()));
    }
    
    <T> void addDecodingRoute(String msgClassString, Adapter<ByteBuffer,T> decoder){
        EventDecoder<Header,T> eventDecoder = new EventDecoder<Header, T>(decoder);
        Listener<Event<Header,ByteBuffer>> listener = 
                new ListenerChain<Event<Header, ByteBuffer>, Event<Header,T>>(
                        eventDecoder, (DefaultRouter)myEventRouter);
        myDecodingRouter.addRoute(msgClassString, listener);
    }
    
    void removeDecodingRoute(String msgClassString){
        myDecodingRouter.removeRoute(msgClassString);
    }
    
    void addEventRoute(String eventKey, Listener<Event<Header,?>> listener){
        myEventRouter.addRoute(eventKey, listener);
    }
    
    void removeEventRoute(String eventKey){
        myEventRouter.removeRoute(eventKey);
    }
    
    public static class DecoderLink<T> {
        private RoutingDecoderNode myRoutingDecoder;
        private String myDecoderKey;
        private Adapter<ByteBuffer,T> myDecoder;
        
        public DecoderLink(RoutingDecoderNode routingDecoder, 
                String decoderKey, Adapter<ByteBuffer,T> decoder){
            myRoutingDecoder = routingDecoder;
            myDecoderKey = decoderKey;
            myDecoder = decoder;
            link();
        }
        
        public void setRoutingDecoder(RoutingDecoderNode routingDecoder){
            unlink();
            myRoutingDecoder = routingDecoder;
            link();
        }
        
        public void setDecoder(String decoderKey, Adapter<ByteBuffer,T> decoder){
            unlink();
            myDecoderKey = decoderKey;
            myDecoder = decoder;
            link();
        }
        
        public void link(){
            if(myRoutingDecoder == null || 
                    myDecoder == null || myDecoderKey == null){
                return;
            }
            myRoutingDecoder.addDecodingRoute(myDecoderKey, myDecoder);
        }
        
        public void unlink(){
            if(myRoutingDecoder == null || myDecoderKey == null){
                return;
            }
            myRoutingDecoder.removeDecodingRoute(myDecoderKey);
        }
    }
    
    public static class EventRouteLink<T> {
        private RoutingDecoderNode myRoutingDecoder;
        private String myRoutingKey;
        private Listener<Event<Header,T>> myEventRoute;
        
        public EventRouteLink(RoutingDecoderNode routingDecoder, 
                String routingKey, Listener<Event<Header,T>> eventRoute){
            myRoutingDecoder = routingDecoder;
            myRoutingKey = routingKey;
            myEventRoute = eventRoute;
            link();
        }
        
        public void setRoutingDecoder(RoutingDecoderNode routingDecoder){
            unlink();
            myRoutingDecoder = routingDecoder;
            link();
        }
        
        public void setRoute(String routingKey, Listener<Event<Header,T>> eventRoute){
            unlink();
            myRoutingKey = routingKey;
            myEventRoute = eventRoute;
            link();
        }
        
        public void link(){
            if(myRoutingDecoder == null || 
                    myEventRoute == null || myRoutingKey == null){
                return;
            }
            myRoutingDecoder.addEventRoute(myRoutingKey, (Listener)myEventRoute);
        }
        
        public void unlink(){
            if(myRoutingDecoder == null || myRoutingKey == null){
                return;
            }
            myRoutingDecoder.removeEventRoute(myRoutingKey);
        }
    }
}
