/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication.chain;

import org.jflux.api.core.Adapter;
import org.jflux.api.core.Listener;
import org.jflux.api.core.Notifier;
import org.jflux.api.core.node.ConsumerNode;
import org.jflux.api.core.node.Node;
import org.jflux.api.core.node.ProcessorNode;
import org.jflux.api.core.node.ProducerNode;

/**
 *
 * @author Matthew Stevenson
 */
public class ChainLinkUtils {    
    public static <T> NodeChainLink<T,ProducerNode<T>,ProcessorNode<T,?>>
            producerProcessorLink(ProducerNode<T> producer, ProcessorNode<T,?> processor){
        return new NodeChainLink<T, ProducerNode<T>, ProcessorNode<T, ?>>(
                producer, processor, 
                new ProducerNotifierAdapter<T>(), 
                new ProcessorListenerAdapter<T>());
    }
    
    public static <T> NodeChainLink<T,ProducerNode<T>,ConsumerNode<T>>
            producerConsumerLink(ProducerNode<T> producer, ConsumerNode<T> consumer){
        return new NodeChainLink<T, ProducerNode<T>, ConsumerNode<T>>(
                producer, consumer, 
                new ProducerNotifierAdapter<T>(), 
                new ConsumerListenerAdapter<T>());
    }
    
    public static <T> NodeChainLink<T,ProcessorNode<?,T>,ConsumerNode<T>>
            processorConsumerLink(ProcessorNode<?,T> processor, ConsumerNode<T> consumer){
        return new NodeChainLink<T, ProcessorNode<?,T>, ConsumerNode<T>>(
                processor, consumer, 
                new ProcessorNotifierAdapter<T>(), 
                new ConsumerListenerAdapter<T>());
    }
    
    public static <T> NodeChainLink<T,ProcessorNode<?,T>,ProcessorNode<T,?>>
            processorLink(ProcessorNode<?,T> procA, ProcessorNode<T,?> procB){
        return new NodeChainLink<T, ProcessorNode<?,T>, ProcessorNode<T,?>>(
                procA, procB, 
                new ProcessorNotifierAdapter<T>(), 
                new ProcessorListenerAdapter<T>());
    }
    
    public static class NodeChainLink<T, A extends Node, B extends Node> 
            extends ChainLink<T> {
        private A myNotifyingNode;
        private B myListeningNode;
        private Adapter<A,Notifier<T>> myAdapterA;
        private Adapter<B,Listener<T>> myAdapterB;

        NodeChainLink(A notifying, B listening,
                Adapter<A,Notifier<T>> adapterA, 
                Adapter<B,Listener<T>> adapterB) {
            super(adapterA.adapt(notifying), adapterB.adapt(listening));
            myNotifyingNode = notifying;
            myListeningNode = listening;
            myAdapterA = adapterA;
            myAdapterB = adapterB;
        }

        public void setNotifyingNode(A notifying) {
            myNotifyingNode = notifying;
            setNotifier(myAdapterA.adapt(myNotifyingNode));
        }

        public void setListeningNode(B listening) {
            myListeningNode = listening;
            setListener(myAdapterB.adapt(myListeningNode));
        }
    }
    
    public static class ProducerNotifierAdapter<T> implements 
            Adapter<ProducerNode<T>,Notifier<T>> {
        @Override public Notifier<T> adapt(ProducerNode<T> a) {
            return a == null ? null : a.getNotifier();
        }
    }
    public static class ProcessorNotifierAdapter<T> implements 
            Adapter<ProcessorNode<?,T>,Notifier<T>> {
        @Override public Notifier<T> adapt(ProcessorNode<?,T> a) {
            return a == null ? null : a.getNotifier();
        }
    }
    public static class ProcessorListenerAdapter<T> implements 
            Adapter<ProcessorNode<T,?>,Listener<T>> {
        @Override public Listener<T> adapt(ProcessorNode<T,?> a) {
            return a == null ? null : a.getListener();
        }
    }
    public static class ConsumerListenerAdapter<T> implements 
            Adapter<ConsumerNode<T>,Listener<T>> {
        @Override public Listener<T> adapt(ConsumerNode<T> a) {
            return a == null ? null : a.getListener();
        }
    }
    
    public static class ChainLink<T> {
        private Notifier<T> myNotifier;
        private Listener<T> myListener;

        public ChainLink(Notifier<T> notifier, Listener<T> listener) {
            myNotifier = notifier;
            myListener = listener;
            addListener();
        }

        public void setNotifier(Notifier<T> notifier) {
            unlink();
            myNotifier = notifier;
            link();
        }

        public void setListener(Listener<T> listener) {
            unlink();
            myListener = listener;
            link();
        }
        
        protected void link(){
            addListener();
        }
        
        private void addListener(){
            if(myListener == null || myNotifier == null){
                return;
            }
            myNotifier.addListener(myListener);
        }
        
        protected void unlink(){
            removeListener();
        }
        
        private void removeListener(){
            if(myListener == null || myNotifier == null){
                return;
            }
            myNotifier.removeListener(myListener);
        }
    }
}
