/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication.chain;

import org.jflux.api.core.node.ProducerNode;
import org.robokind.demo.robot.replication.chain.ChainLinkUtils.ProducerNotifierAdapter;
import org.robokind.demo.robot.replication.chain.ChainLinkUtils.ConsumerListenerAdapter;
import org.jflux.api.core.node.ConsumerNode;
import org.robokind.demo.robot.replication.chain.ChainLinkUtils.ProcessorListenerAdapter;
import org.robokind.demo.robot.replication.chain.ChainLinkUtils.ProcessorNotifierAdapter;
import org.jflux.api.core.Notifier;
import org.robokind.demo.robot.replication.chain.ChainLinkUtils.NodeChainLink;
import org.jflux.api.core.node.Node;
import org.jflux.api.core.node.ProcessorNode;
import java.util.Arrays;
import java.util.Map;
import java.util.Properties;
import org.jflux.api.core.Adapter;
import org.jflux.api.core.Listener;
import org.jflux.api.core.config.Configuration;
import org.robokind.api.common.lifecycle.config.GenericLifecycle.DependencyChange;
import static org.robokind.api.common.lifecycle.config.RKDependencyConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKLifecycleConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKManagedGroupConfigUtils.*;

/**
 *
 * @author Matthew Stevenson
 */
public class NodeLinkLifecycleUtils {
    private final static String theNotifyingNode = "encodeConfig";    
    private final static String theListeningNode = "destination";
    
    public final static String PROP_NODE_LINK_ID = "nodeLinkId";
    
    public static <T> Configuration<String> producerConsumerLinkLifecycleConfig(
                    String id, Properties linkProps,
                    String notifyingNodeIdKey, String notifyingNodeId, 
                    Properties notifyingProps, 
                    String listeningNodeIdKey, String listeningNodeId,
                    Properties listeningProps){
        return nodeLinkLifecycleConfig(PROP_NODE_LINK_ID, id, linkProps, 
                ProducerNode.class, notifyingNodeIdKey, 
                notifyingNodeId, notifyingProps,
                (Adapter)new ProducerNotifierAdapter<T>(), 
                ConsumerNode.class, listeningNodeIdKey, 
                listeningNodeId, listeningProps,
                (Adapter)new ConsumerListenerAdapter<T>());
    }
    
    public static <T> Configuration<String> producerProcessorLinkLifecycleConfig(
                    String id, Properties linkProps,
                    String notifyingNodeIdKey, String notifyingNodeId, 
                    Properties notifyingProps, 
                    String listeningNodeIdKey, String listeningNodeId,
                    Properties listeningProps){
        return nodeLinkLifecycleConfig(PROP_NODE_LINK_ID, id, linkProps, 
                ProducerNode.class, notifyingNodeIdKey, 
                notifyingNodeId, notifyingProps,
                (Adapter)new ProducerNotifierAdapter<T>(), 
                ProcessorNode.class, listeningNodeIdKey, 
                listeningNodeId, listeningProps,
                (Adapter)new ProcessorListenerAdapter<T>());
    }
    
    public static <T> Configuration<String> processorConsumerLinkLifecycleConfig(
                    String id, Properties linkProps,
                    String notifyingNodeIdKey, String notifyingNodeId, 
                    Properties notifyingProps, 
                    String listeningNodeIdKey, String listeningNodeId,
                    Properties listeningProps){
        return nodeLinkLifecycleConfig(PROP_NODE_LINK_ID, id, linkProps, 
                ProcessorNode.class, notifyingNodeIdKey, 
                notifyingNodeId, notifyingProps,
                (Adapter)new ProcessorNotifierAdapter<T>(), 
                ConsumerNode.class, listeningNodeIdKey, 
                listeningNodeId, listeningProps,
                (Adapter)new ConsumerListenerAdapter<T>());
    }
    
    public static <T> Configuration<String> processorLinkLifecycleConfig(
                    String id, Properties linkProps,
                    String notifyingNodeIdKey, String notifyingNodeId, 
                    Properties notifyingProps, 
                    String listeningNodeIdKey, String listeningNodeId,
                    Properties listeningProps){
        return nodeLinkLifecycleConfig(PROP_NODE_LINK_ID, id, linkProps, 
                ProcessorNode.class, notifyingNodeIdKey, 
                notifyingNodeId, notifyingProps,
                (Adapter)new ProcessorNotifierAdapter<T>(), 
                ProcessorNode.class, listeningNodeIdKey, 
                listeningNodeId, listeningProps,
                (Adapter)new ProcessorListenerAdapter<T>());
    }
    
    public static <T, A extends Node, B extends Node> 
            Configuration<String> nodeLinkLifecycleConfig(
                    String idKey, String id, Properties linkProps,
                    Class<A> notifyingNodeClass, String notifyingNodeIdKey, 
                    String notifyingNodeId, Properties notifyingProps, 
                    Adapter<A,Notifier<T>> adapterA, 
                    Class<B> listeningNodeClass, String listeningNodeIdKey, 
                    String listeningNodeId, Properties listeningProps, 
                    Adapter<B,Listener<T>> adapterB){
        return buildNodeLifecycleConfig(idKey, id, 
                buildNotifyingDepConfig(notifyingNodeClass, notifyingNodeIdKey, 
                        notifyingNodeId, notifyingProps), 
                buildListeningDepConfig(listeningNodeClass, listeningNodeIdKey, 
                        listeningNodeId, listeningProps), 
                linkProps, adapterA, adapterB);
    }
    
    public static <T, A extends Node, B extends Node> 
            Configuration<String> buildNodeLifecycleConfig(
                    String linkIdKey, String linkId,
                    Configuration<String> notifyingConfig, 
                    Configuration<String> listeningConfig, 
                    Properties linkProperties,
                    final Adapter<A,Notifier<T>> adapterA, 
                    final Adapter<B,Listener<T>> adapterB){
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{NodeChainLink.class.getName()}, 
                linkIdKey, linkId, linkProperties, 
                Arrays.asList(notifyingConfig,listeningConfig), 
                new Adapter<Map<String,Object>,NodeChainLink>() {
                    @Override
                    public NodeChainLink adapt(Map<String,Object> a) {
                        return new NodeChainLink<T, A, B>(
                                (A)a.get(theNotifyingNode), 
                                (B)a.get(theListeningNode), 
                                adapterA, adapterB);
                    }},
                new Listener<NodeChainLink>() {
                    @Override public void handleEvent(NodeChainLink input) {
                        input.setListeningNode(null);
                        input.setNotifyingNode(null);
                    }
                }));
    }
    
    public static <T, N extends Node> 
            Configuration<String> buildNotifyingDepConfig(Class<N> nodeClass, 
                    String nodeIdKey, String nodeId, Properties props){
        return buildLifecycleDependencyConfig(
                theNotifyingNode, nodeClass, nodeIdKey, nodeId, props, 
                new Listener<DependencyChange<NodeChainLink<T,N,?>,N>>(){
                    @Override public void handleEvent(
                            DependencyChange<NodeChainLink<T,N,?>,N> input) {
                        NodeChainLink<T,N,?> link = input.getService();
                        if(link == null){
                            return;
                        }
                        link.setNotifyingNode(input.getDependency());
                    }});
    }
    
    public static <T, N extends Node> 
            Configuration<String> buildListeningDepConfig(Class<N> nodeClass, 
                    String nodeIdKey, String nodeId, Properties props){
        return buildLifecycleDependencyConfig(
                theListeningNode, nodeClass, nodeIdKey, nodeId, props, 
                new Listener<DependencyChange<NodeChainLink<T,?,N>,N>>(){
                    @Override public void handleEvent(
                            DependencyChange<NodeChainLink<T,?,N>,N> input) {
                        NodeChainLink<T,?,N> link = input.getService();
                        if(link == null){
                            return;
                        }
                        link.setListeningNode(input.getDependency());
                    }});
    }
}
