/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.robokind.demo.robot.replication.lifecycle;

import org.jflux.api.core.Listener;
import org.jflux.api.core.node.ConsumerNode;
import java.util.Properties;
import java.util.List;
import org.jflux.api.core.Source;
import org.jflux.api.core.chain.ListenerChain;
import org.jflux.api.core.chain.SourceChain;
import org.jflux.api.core.util.BatchAdapter;
import org.jflux.api.data.buffer.Buffer;
import org.robokind.demo.robot.replication.EventEncoder.HeaderRecordSource;
import java.nio.ByteBuffer;
import org.apache.avro.Schema;
import java.io.ByteArrayOutputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.apache.avro.generic.IndexedRecord;
import org.jflux.api.core.Adapter;
import org.jflux.api.core.chain.AdapterChain;
import org.jflux.api.core.config.Configuration;
import org.jflux.api.core.event.BasicEvent.BasicEventFactory;
import org.jflux.api.core.event.Event;
import org.jflux.api.core.event.Header;
import org.jflux.api.core.node.DefaultConsumerNode;
import org.jflux.api.core.util.EmptyListener;
import org.jflux.api.core.util.ListAdapter;
import org.jflux.api.core.util.MapAdapter.MapValueAdapter;
import org.jflux.api.core.util.SourceAdapter;
import org.jflux.api.data.buffer.CircularBuffer;
import org.jflux.api.encode.BytesUtils;
import org.jflux.api.encode.EncodeRequest;
import org.jflux.avrogen.EventRecord;
import org.jflux.impl.encode.avro.AvroEncoder;
import org.jflux.impl.messaging.JMSAvroUtils.ByteOutputStreamFactory;
import org.robokind.demo.robot.replication.EventEncoder;
import static org.jflux.impl.encode.avro.SerializationConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKDependencyConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKLifecycleConfigUtils.*;
import static org.robokind.api.common.lifecycle.config.RKManagedGroupConfigUtils.*;

/**
 *
 * @author Matthew Stevenson
 */
public class EventBufferUtils {
    private final static String theEncodeConfig = "encodeConfig";
    private final static String theDataEventFactory = "dataEventFactory";
    private final static String theEventEncoder = "eventEncoder";
    
    private final static String theDataEventFactoryIdKey = "encodeConfig";
    private final static String theEventEncoderIdKey = "encodeConfig";
    
    public final static String PROP_EVENT_BUFFER_NODE_ID = "eventBufferNode";
    
    public final static String PROP_EVENT_TYPE = "eventType";
    
    public static List<Configuration<String>> getLifecycleConfigs(
            String encodeConfigIdKey, String encodeConfigId, 
            String eventFactoryId, String eventKey, 
            String sourceRef, String eventEncoderId, 
            String eventBufferId, Properties bufferProps, int capacity){
        return Arrays.asList(
            buildDataEventFactoryLifecycleConfig(encodeConfigIdKey, encodeConfigId, 
                eventFactoryId, eventKey, sourceRef),
            buildEventEncoderLifecycleConfig(
                encodeConfigIdKey, encodeConfigId, eventEncoderId),
            buildEventBufferLifecycleConfig(
                eventBufferId, bufferProps, 
                capacity, eventFactoryId, eventEncoderId));
    }
    
    public static Configuration<String> buildDataEventFactoryLifecycleConfig(
            String encodeConfigIdKey, String encodeConfigId, String eventFactoryId, 
            final String eventKey, final String sourceRef){
        Configuration<String> depConf = buildLifecycleDependencyConfig(
                theEncodeConfig, Configuration.class, 
                encodeConfigIdKey, encodeConfigId, null, null);
        Properties props = new Properties();
        props.put(PROP_EVENT_TYPE, eventKey);
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{Adapter.class.getName(), BasicEventFactory.class.getName()}, 
                theDataEventFactoryIdKey, eventFactoryId, props, Arrays.asList(depConf), 
                new MapValueAdapter(theEncodeConfig, new Adapter<Configuration,BasicEventFactory>() {
                    @Override public BasicEventFactory adapt(Configuration a) {
                        return new BasicEventFactory(
                                buildHeaderSource(eventKey, sourceRef, a));
                    }})));
    }
    
    private static HeaderRecordSource buildHeaderSource(
            String eventKey, String sourceRef, Configuration<String> conf){
        Class msgClass = conf.getPropertyValue(CONF_MESSAGE_CLASS);
        Map<String,String> props = new HashMap<String, String>(1);
        props.put(CONF_MESSAGE_CLASS, msgClass.toString());
        return new HeaderRecordSource(eventKey, sourceRef, props);
    }
    
    public static <T, R extends IndexedRecord> Configuration<String> buildEventEncoderLifecycleConfig(
            String encodeConfigIdKey, String encodeConfigId, String eventEncoderId){
        Configuration<String> depConf = buildLifecycleDependencyConfig(
                theEncodeConfig, Configuration.class, 
                encodeConfigIdKey, encodeConfigId, null, null);
        
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{Adapter.class.getName(), EventEncoder.class.getName()}, 
                theEventEncoderIdKey, eventEncoderId, null, Arrays.asList(depConf), 
                new MapValueAdapter(theEncodeConfig, new Adapter<Configuration,EventEncoder>() {
                    @Override
                    public EventEncoder adapt(Configuration a) {
                        return new EventEncoder(encoder(a));
                    }})));
    }
    
    private static <T,R extends IndexedRecord> Adapter<T,ByteBuffer> encoder(
            Configuration<String> config){
        Adapter<T,R> msgAdapter = config.getPropertyValue(CONF_ENCODING_ADAPTER);
        Class<R> clazz = config.getPropertyValue(CONF_OUTPUT_CLASS);
        Schema schema = config.getPropertyValue(CONF_AVRO_RECORD_SCHEMA);        
        return AdapterChain.builder(msgAdapter)
                .attach(EncodeRequest.factory(clazz, new ByteOutputStreamFactory()))
                .attach(new AvroEncoder<R, ByteArrayOutputStream>(clazz, schema, false))
                .attach(BytesUtils.outputStreamToBuffer()).done();
    }
    
    public static <T> Configuration<String> buildEventBufferLifecycleConfig(
            final String eventBufferId, Properties bufferProps, 
            final int capacity, String eventFactId, String eventEncoderId){
        List<Configuration<String>> deps = Arrays.asList(
            buildLifecycleDependencyConfig(theDataEventFactory, BasicEventFactory.class, 
                theDataEventFactoryIdKey, eventFactId, null, null),
            buildLifecycleDependencyConfig(theEventEncoder, EventEncoder.class, 
                theEventEncoderIdKey, eventEncoderId, null, null)
        );
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{EventBuffer.class.getName(), ConsumerNode.class.getName()}, 
                PROP_EVENT_BUFFER_NODE_ID, eventBufferId, null, deps, 
                new Adapter<Map<String, Object>, EventBuffer>() {
                    @Override public EventBuffer adapt(Map<String, Object> a) {
                        BasicEventFactory eventFactory = (BasicEventFactory)a.get(theDataEventFactory);
                        EventEncoder eventEncoder = (EventEncoder)a.get(theEventEncoder);
                        Buffer<Integer,Event<Header<String,Long>,T>> b = 
                                new CircularBuffer<Event<Header<String,Long>,T>>(capacity);
                        EventBuffer eb = 
                                new EventBuffer(eventFactory, b, eventEncoder);
                        eb.start();
                        return eb;
                    }
                },
                new Listener<EventBuffer>() {
                    @Override public void handleEvent(EventBuffer input) {
                        input.stop();
                    }
                }));        
    }
    
    
    public static List<Configuration<String>> getAlternateLifecycleConfigs(
            String encodeConfigIdKey, String encodeConfigId, 
            String eventFactoryId, String eventKey, 
            String sourceRef, String eventEncoderId, 
            String eventBufferId, Source<?> valueSource){
        return Arrays.asList(
            buildDataEventFactoryLifecycleConfig(encodeConfigIdKey, encodeConfigId, 
                eventFactoryId, eventKey, sourceRef),
            buildEventEncoderLifecycleConfig(
                encodeConfigIdKey, encodeConfigId, eventEncoderId),
            buildAlternateEventBufferLifecycleConfig(
                eventBufferId, eventFactoryId, eventEncoderId, 
                valueSource));
    }
    
    public static <T> Configuration<String> buildAlternateEventBufferLifecycleConfig(
            final String eventBufferId, String eventFactId, 
            String eventEncoderId, final Source<T> valueSource){
        List<Configuration<String>> deps = Arrays.asList(
            buildLifecycleDependencyConfig(theDataEventFactory, BasicEventFactory.class, 
                theDataEventFactoryIdKey, eventFactId, null, null),
            buildLifecycleDependencyConfig(theEventEncoder, EventEncoder.class, 
                theEventEncoderIdKey, eventEncoderId, null, null)
        );
        return makeSelfBuildingLifecycle(buildGenericLifecycleConfig(
                new String[]{EventBuffer.class.getName(), ConsumerNode.class.getName()}, 
                PROP_EVENT_BUFFER_NODE_ID, eventBufferId, null, deps, 
                new Adapter<Map<String, Object>, EventBuffer>() {
                    @Override public EventBuffer adapt(Map<String, Object> a) {
                        BasicEventFactory eventFactory = (BasicEventFactory)a.get(theDataEventFactory);
                        EventEncoder eventEncoder = (EventEncoder)a.get(theEventEncoder);
                        Buffer<Integer,Event<Header<String,Long>,T>> b = 
                                new SourceBuffer<T>(valueSource, eventFactory);
                        EventBuffer eb = new EventBuffer(eventFactory, b, eventEncoder);
                        eb.start();
                        return eb;
                    }
                },
                new Listener<EventBuffer>() {
                    @Override public void handleEvent(EventBuffer input) {
                        input.stop();
                    }
                }));        
    }
    
    public static class EventBuffer<T> extends DefaultConsumerNode<T> {
        private Buffer<Integer,Event<Header<String,Long>,T>> myBuffer;
        private Adapter<Event<Header<String,Long>,T>,EventRecord> myEventAdapter;
        private Source<List<EventRecord>> myRecordSource;

        public EventBuffer(
                Adapter<T,Event<Header<String,Long>,T>> eventFactory,
                Buffer<Integer,Event<Header<String,Long>,T>> buffer,
                Adapter<Event<Header<String,Long>,T>,EventRecord> eventAdapter){
            super(new ListenerChain<T, Event<Header<String,Long>,T>>(
                    eventFactory, buffer.addValue()));
            if(eventFactory == null || buffer == null || eventAdapter == null){
                throw new NullPointerException();
            }
            myBuffer = buffer;
            myEventAdapter = eventAdapter;
            myRecordSource = 
                    new SourceChain<Iterable<Event<Header<String,Long>,T>>,List<EventRecord>>(
                            (Source)myBuffer.getValues(), new BatchAdapter(myEventAdapter));
        }

        public Source<List<EventRecord>> getRecordValues() {
            return myRecordSource;
        }
    }
    
    
    static class SourceBuffer<V> implements Buffer<Integer, Event<Header<String,Long>,V>> {
        private Source<V> myValueSource;
        private Adapter<V, Event<Header<String,Long>,V>> myEventAdapter;
        private Source<Event<Header<String, Long>, V>> myEventSource;
        private Source<? extends Iterable<Event<Header<String, Long>, V>>> myValuesSource;
        private Adapter<Integer, Event<Header<String, Long>, V>> myIndexAdapter;
        private Listener<Event<Header<String, Long>, V>> myValueListener;

        public SourceBuffer(
                Source<V> valueSource, 
                Adapter<V, Event<Header<String, Long>, V>> eventAdapter) {
            myValueSource = valueSource;
            myEventAdapter = eventAdapter;
            myEventSource = new SourceChain<V, Event<Header<String, Long>, V>>(
                    myValueSource, myEventAdapter);
            myValuesSource = new SourceChain<Event<Header<String, Long>, V>, 
                    List<Event<Header<String, Long>, V>>>(
                            myEventSource, new ListAdapter(1));
            myIndexAdapter = new SourceAdapter<
                    Integer, Event<Header<String, Long>, V>>(myEventSource);
            myValueListener = new EmptyListener<Event<Header<String, Long>, V>>();
        }
        
        @Override
        public Source<Event<Header<String, Long>, V>> getHead() {
            return myEventSource;
        }

        @Override
        public Source<Event<Header<String, Long>, V>> getTail() {
            return myEventSource;
        }

        @Override
        public Adapter<Integer, Event<Header<String, Long>, V>> getIndex() {
            return myIndexAdapter;
        }

        @Override
        public Listener<Event<Header<String, Long>, V>> addValue() {
            return myValueListener;
        }

        @Override
        public Source<? extends Iterable<Event<Header<String, Long>, V>>> getValues() {
            return myValuesSource;
        }
    }
}
